/*
 * Copyright (C) 1997-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afh_common.c Common audio format handler functions. */

#include <sys/mman.h> /* mmap */
#include <sys/time.h> /* gettimeofday */
#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"

/* The mp3 audio format handler does not need any libs. */
void mp3_init(struct audio_format_handler *);

#ifdef HAVE_OGGVORBIS
	void ogg_init(struct audio_format_handler *);
#endif
#ifdef HAVE_FAAD
	void aac_afh_init(struct audio_format_handler *);
#endif

/**
 * The list of supported audio formats.
 *
 * We always define the full array of audio formats even if some audio formats
 * were not compiled in. This is because for each audio file the number of its
 * audio format is stored in the database. We don't want that numbers to become
 * stale just because the user installed a new version of paraslash that
 * supports a different set of audio formats.
 *
 * It can still be easily detected whether an audio format is compiled in by
 * checking if the init function pointer is not \p NULL.
 */
static struct audio_format_handler afl[] = {
	{
		.name = "mp3",
		.init = mp3_init,
	},
	{
		.name = "ogg",
#ifdef HAVE_OGGVORBIS
		.init = ogg_init,
#endif
	},
	{
		.name = "aac",
#ifdef HAVE_FAAD
		.init = aac_afh_init,
#endif
	},
	{
		.name = NULL,
	}
};

static inline int next_audio_format(int format)
{
	for (;;) {
		if (!afl[format].name)
			return format;
		format++;
		if (afl[format].init)
			return format;
	}

}

/** Iterate over each supported audio format. */
#define FOR_EACH_AUDIO_FORMAT(i) for (i = 0; afl[i].name; i = next_audio_format(i))

/**
 * Call the init function of each supported audio format handler.
 */
void afh_init(void)
{
	int i;

	PARA_INFO_LOG("supported audio formats: %s\n",
		SUPPORTED_AUDIO_FORMATS);
	FOR_EACH_AUDIO_FORMAT(i) {
		PARA_NOTICE_LOG("initializing %s handler\n",
			audio_format_name(i));
		afl[i].init(&afl[i]);
	}
}

/**
 * Guess the audio format judging from filename.
 *
 * \param name The filename.
 *
 * \return This function returns \p -E_AUDIO_FORMAT if it has no idea what kind
 * of audio file this might be. Otherwise the (non-negative) number of the
 * audio format is returned.
 */
int guess_audio_format(const char *name)
{
	int i,j, len = strlen(name);

	FOR_EACH_AUDIO_FORMAT(i) {
		for (j = 0; afl[i].suffixes[j]; j++) {
			const char *p = afl[i].suffixes[j];
			int plen = strlen(p);
			if (len < plen + 1)
				continue;
			if (name[len - plen - 1] != '.')
				continue;
			if (strcasecmp(name + len - plen, p))
				continue;
//			PARA_DEBUG_LOG("might be %s\n", audio_format_name(i));
			return i;
		}
	}
	return -E_AUDIO_FORMAT;
}

/**
 * Pretty-print the given meta-info.
 *
 * \param title The title of the audio file.
 * \param artist The artist.
 * \param album The name of the album.
 * \param year Year of release.
 * \param comment Further comments.
 *
 * This function is called by each audio format handler to produce the tag info
 * status items. Usually, the audio format handlers read this info from the
 * audio file (id3 tags, vorbis comments, ...).
 *
 * It is OK to pass \p NULL pointers for any argument in which case a suitable
 * string is inserted which indicates that this information is not available.
 *
 * \return The status item string. It must be freed by the caller.
 */
char *make_taginfo(char *title, char *artist, char *album, char *year,
		char *comment)
{
	return make_message("%s: %s, by %s\n" /* taginfo1 */
		"%s: A: %s, Y: %s, C: %s\n", /* taginfo2 */
		status_item_list[SI_TAGINFO1],
		(title && *title)? title : "(title tag not set)",
		(artist && *artist)? artist : "(artist tag not set)",
		status_item_list[SI_TAGINFO2],
		(album && *album)?  album : "(album tag not set)",
		(year && *year)? year : "????",
		(comment && *comment)? comment : "(comment tag not set)"
	);
}

/**
 * Call get_file_info() to obtain an afhi structure.
 *
 * \param path The full path of the audio file.
 * \param data Pointer to the contents of the (mapped) file.
 * \param size The file size in bytes.
 * \param fd The open file descriptor.
 * \param afhi Result pointer.
 *
 * \return The number of the audio format on success, \p -E_AUDIO_FORMAT if no
 * compiled in audio format handler is able to handler the file.
 *
 * This function tries to find an audio format handler that can interpret the
 * file given by \a data and \a size.
 *
 * It first tries to determine the audio format from the filename given by \a
 * path. If this doesn't work, all other audio format handlers are tried until
 * one is found that can handle the file.
 */
int compute_afhi(const char *path, char *data, size_t size, int fd,
		struct afh_info *afhi)
{
	int ret, i, format;

	afhi->header_offset = 0;
	afhi->header_len = 0;
	format = guess_audio_format(path);

	if (format >= 0) {
		ret = afl[format].get_file_info(data, size, fd, afhi);
		if (ret >= 0)
			return format;
	}
	FOR_EACH_AUDIO_FORMAT(i) {
		if (i == format) /* we already tried this one to no avail */
			continue;
		ret = afl[i].get_file_info(data, size, fd, afhi);
		if (ret >= 0)
			return i;
		PARA_WARNING_LOG("%s\n", para_strerror(-ret));
	}
	return -E_AUDIO_FORMAT;
}

/**
 * Get the name of the given audio format.
 *
 * \param i The audio format number.
 *
 * This returns a pointer to statically allocated memory so it
 * must not be freed by the caller.
 */
const char *audio_format_name(int i)
{
	//PARA_NOTICE_LOG("array size: %u requested: %d\n", ARRAY_SIZE(afl), i);
	assert(i < 0 || i < ARRAY_SIZE(afl) - 1);
	return i >= 0?  afl[i].name : "(none)";
}

/**
 * Get one chunk of audio data.
 *
 * \param chunk_num The number of the chunk to get.
 * \param afhi Describes the audio file.
 * \param map The memory mapped audio file.
 * \param buf Result pointer.
 * \param len The length of the chunk in bytes.
 *
 * Upon return, \a buf will point so memory inside \a map. The returned buffer
 * must therefore not be freed by the caller.
 */
void afh_get_chunk(long unsigned chunk_num, struct afh_info *afhi,
		void *map, const char **buf, size_t *len)
{
	size_t pos = afhi->chunk_table[chunk_num];
	*buf = map + pos;
	*len = afhi->chunk_table[chunk_num + 1] - pos;
}

uint32_t afh_get_largest_chunk_size(struct afh_info *afhi)
{
	uint32_t n, largest = 0, *ct = afhi->chunk_table;

	for (n = 1; n <= afhi->chunks_total; n++)
		largest = PARA_MAX(largest, ct[n] - ct[n - 1]);
	return largest;
}

/**
 * Get the header of an audio file.
 *
 * \param afhi The audio file handler data describing the file.
 * \param map The data of the audio file.
 * \param buf The length of the header is stored here.
 * \param len Points to a buffer containing the header on return.
 *
 * This function sets \a buf to \p NULL and \a len to zero if \a map or \a
 * afhi is \p NULL, or if the current audio format does not need special
 * header treamtment.
 */
void afh_get_header(struct afh_info *afhi, void *map, const char **buf, size_t *len)
{
	if (!map || !afhi || !afhi->header_len) {
		*buf = NULL;
		*len = 0;
		return;
	}
	*len = afhi->header_len;
	*buf = map + afhi->header_offset;
}
