/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file recv.c the stand-alone audio stream receiver */

#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "recv.cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "stdout.h"

/** the gengetopt args info struct */
struct recv_args_info conf;

static int loglevel;
/** Always log to stderr. */
INIT_STDERR_LOGGING(loglevel);

/** init array of error codes used by para_recv */
INIT_RECV_ERRLISTS;

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? recv_args_info_detailed_help
		: recv_args_info_help;

	printf_or_die("%s\n\n", RECV_CMDLINE_PARSER_PACKAGE "-"
		RECV_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", recv_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_receiver_helps(d);
	exit(0);
}

static void *parse_config(int argc, char *argv[], int *receiver_num)
{
	if (recv_cmdline_parser(argc, argv, &conf))
		return NULL;
	HANDLE_VERSION_FLAG("recv", conf);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	return check_receiver_arg(conf.receiver_arg, receiver_num);
}

/**
 * the main function of para_recv
 *
 * \param argc number of arguments
 * \param argv vector of arguments
 *
 * para_recv uses the specified receiver to receive an audio stream sent by
 * para_server. The received data is written to stdout.
 *
 * \return \a EXIT_SUCCESS on success, \a EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	int ret, r_opened = 0, receiver_num;
	struct receiver *r = NULL;
	struct receiver_node rn;
	struct stdout_task sot;
	static struct sched s;

	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;

	memset(&sot, 0, sizeof(struct stdout_task));
	memset(&rn, 0, sizeof(struct receiver_node));
	recv_init();
	ret = -E_RECV_SYNTAX;
	rn.conf = parse_config(argc, argv, &receiver_num);
	if (!rn.conf) {
		PARA_EMERG_LOG("parse failed\n");
		goto out;
	}
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	r = &receivers[receiver_num];
	rn.receiver = r;
	ret = r->open(&rn);
	if (ret < 0)
		goto out;
	r_opened = 1;

	stdout_set_defaults(&sot);
	sot.buf = rn.buf;
	sot.loaded = &rn.loaded;
	sot.input_error = &rn.task.error;
	register_task(&sot.task);

	rn.task.pre_select = r->pre_select;
	rn.task.post_select = r->post_select;
	sprintf(rn.task.status, "receiver node");
	register_task(&rn.task);

	ret = schedule(&s);
out:
	if (r_opened)
		r->close(&rn);
	if (r)
		r->shutdown();
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
